
#include "utils.h"

#include <errno.h>
#include <string.h>

#define GPIOCOUNT 20
#define PINCOUNT 40

#define EXPORT_PATH "/sys/class/gpio/export"
#define UNEXPORT_PATH "/sys/class/gpio/unexport"

#define PINS -1, -1, -1, -1, -1, -1, 216, -1, -1, -1, 50, 79, 14, -1, 194, 232, -1, 15, 16, -1, 17, 13, 18, 19, -1, 20, -1, -1, 149, -1, 200, 168, 38, -1, 76, 51, 12, 77, -1, 78

#define POLL_GPIO POLLPRI | POLLERR

struct GPIO *gpio_pins[40];
int gpios_list[40] = {PINS};

void fatal_error(int error_code, char *msg)
{
  printf("Error code %d : %s \n", error_code, msg);
  destroy_gpios();
  exit(-1);
};

void check_pin(int pin)
{
  if (gpios_list[pin - 1] == -1)
    fatal_error(33, "Le numéro de Pin spécifié est incrrect");
}

int is_gpio(int pin)
{
  return gpios_list[pin - 1] != -1;
}

void destroy_gpios()
{
  for (int i = 0; i < 40; i++)
    if (is_gpio(i + 1))
      unpin_mode(i);
}
struct GPIO *load_pin(int pin)
{
  struct GPIO *gpio = (struct GPIO *)(malloc(sizeof(struct GPIO)));
  gpio->pin = pin + 1;
  gpio->gpio = gpios_list[pin];
  gpio->value = -1;
  gpio->direction = -1;
  gpio->fd_direction = -1;
  gpio->fd_value = -1;
  sprintf(gpio->gpio_s, "%d", gpio->gpio);
  sprintf(gpio->str_direction, "/sys/class/gpio/gpio%d/direction", gpio->gpio);
  sprintf(gpio->str_value, "/sys/class/gpio/gpio%d/value", gpio->gpio);
  return gpio;
}

void init_GPIOs()
{
  for (int i = 0; i < 40; i++)
  {
    if (gpios_list[i] == -1)
      gpio_pins[i] = NULL;
    else
    {
      gpio_pins[i] = load_pin(i);
    }
  }
};

void digital_write(int pin, char l)
{
  char ll = l;
  write(gpio_pins[pin - 1]->fd_value, &l, 1);
}

int digital_read(int pin)
{
  char c;
  pread(gpio_pins[pin - 1]->fd_value, &c, 1, 0);
  return c == '1';
}

int pin_mode(int pin, int direction)
{
  check_pin(pin);
  if (!gpio_pins[pin - 1]->locked)
  {
    export_pin(gpio_pins[pin - 1]);
  }
  set_direction(gpio_pins[pin - 1], direction);
  open_value(gpio_pins[pin - 1]);
}

void export_pin(struct GPIO *pin)
{
  printf("%s \n", EXPORT_PATH);
  int fd = open(EXPORT_PATH, O_WRONLY);
  if (fd == -1)
    fatal_error(2, "Something went wrong \n");
  write(fd, pin->gpio_s, 3);
  pin->locked = 1;
  close(fd);
}

void unpin_mode(int pin)
{
  close(gpio_pins[pin - 1]->fd_value);
  close(gpio_pins[pin - 1]->direction);
  int fd = open(UNEXPORT_PATH, O_WRONLY);
  if (fd == -1)
    fatal_error(2, "Something went wrong \n");
  write(fd, gpio_pins[pin - 1]->gpio_s, 3);
  gpio_pins[pin - 1]->locked = 0;
  close(fd);
}

void set_direction(struct GPIO *pin, int direction)
{
  int fd = open(pin->str_direction, O_WRONLY);
  if (fd == -1)
    fatal_error(2, "Something went wrong \n");
  if (direction == OUTPUT)
    write(fd, "out", 3);
  else
    write(fd, "in", 3);
  pin->direction = direction;
  pin->fd_direction = fd;

  // Setting Polling params
  if (direction == INPUT)
  {
    pin->poll_gpio.fd = pin->fd_value;
    pin->poll_gpio.events = POLL_GPIO;
    pin->poll_gpio.revents = 0;
  }
}

void open_value(struct GPIO *pin)
{
  int fd = open(pin->str_value, O_RDWR);
  if (fd == -1)
    fatal_error(2, "Something went wrong \n");
  printf("called %d", fd);
  pin->fd_value = fd;
  printf("---> %d \n", fd);
}

void display(const struct GPIO *gp)
{
  if (gp == NULL)
    printf("[GPIO : Empty ] \n");
  else
    printf("[GPIO  pin : %d, gpio : %d, gpio_s : %s, fd_v : %d] \n", gp->pin, gp->gpio, gp->gpio_s, gp->fd_value);
}

int wait_for_toggle(int pin, int timeout, void *(func)(void *), void *args)
{
  display(gpio_pins[pin - 1]);
  int poll_ret = poll(&(gpio_pins[pin - 1]->poll_gpio), 1, timeout);
  if (!poll_ret)
  {
    (*func)(args);
    return 0;
  }
  else
  {
    if (poll_ret == -1)
      return 0;
    if ((gpio_pins[pin - 1]->poll_gpio.revents) & (POLL_GPIO))
      return 1;
  }
}
